"""Decorator to monitor devices."""
import asyncio
from collections import defaultdict

from mpf.core.utility_functions import Util


class DeviceMonitor:

    """Monitor variables of a device."""

    __slots__ = ["_attributes_to_monitor", "_aliased_attributes_to_monitor", "_do_not_overwrite_setter"]

    def __init__(self, *attributes_to_monitor, _do_not_overwrite_setter=False, **aliased_attributes_to_monitor):
        """Initialise decorator and remember attributes to monitor."""
        self._attributes_to_monitor = attributes_to_monitor
        self._aliased_attributes_to_monitor = aliased_attributes_to_monitor
        self._do_not_overwrite_setter = _do_not_overwrite_setter

    def __call__(self, cls):    # noqa
        """Decorate class."""
        _sentinel = object()

        old_init = getattr(cls, '__init__', None)

        def __init__(self_inner, *args, **kwargs):  # noqa
            """Register class."""
            old_init(self_inner, *args, **kwargs)
            self_inner.machine.device_manager.register_monitorable_device(self_inner)

        old_setattr = getattr(cls, '__setattr__', None)
        super_get_placeholder_value = getattr(cls, "get_placeholder_value", None)

        def __setattr__(self_inner, name, value):   # noqa
            """If the value changed notify subscribers."""
            attribute_name = False
            if name in self._attributes_to_monitor:
                old = getattr(self_inner, name, _sentinel)
                if old is not _sentinel and old != value:
                    attribute_name = name
            elif name in self._aliased_attributes_to_monitor:
                old = getattr(self_inner, name, _sentinel)
                if old is not _sentinel and old != value:
                    attribute_name = self._aliased_attributes_to_monitor[name]

            if old_setattr:
                old_setattr(self_inner, name, value)
            else:
                # Old-style class
                self_inner.__dict__[name] = value

            if attribute_name:
                _notify_placeholder_change(self_inner, attribute_name, old, value)

        def _notify_placeholder_change(self_inner, attribute_name, old, value):
            if old != value:
                self_inner.machine.device_manager.notify_device_changes(self_inner, attribute_name, old, value)
                for future in cls.attribute_futures[self_inner][attribute_name]:
                    if not future.done():
                        future.set_result(True)
                cls.attribute_futures[self_inner][attribute_name] = []

        def get_monitorable_state(self_inner):
            """Return monitorable state of device."""
            state = {}
            for attribute in self._attributes_to_monitor:
                state[attribute] = Util.convert_to_simply_type(getattr(self_inner, attribute))

            for attribute, name in self._aliased_attributes_to_monitor.items():
                state[name] = Util.convert_to_simply_type(getattr(self_inner, attribute))

            return state

        def subscribe_attribute(self_inner, item, machine):
            """Subscribe to an attribute."""
            del machine
            future = asyncio.Future()
            cls.attribute_futures[self_inner][item].append(future)
            return future

        def get_placeholder_value(self_inner, item):
            """Get the value of a placeholder."""
            if item in self._attributes_to_monitor:
                return Util.convert_to_simply_type(getattr(self_inner, item))

            for attribute, name in self._aliased_attributes_to_monitor.items():
                if name == item:
                    return Util.convert_to_simply_type(getattr(self_inner, attribute))

            if super_get_placeholder_value:
                return super_get_placeholder_value(self_inner, item)

            raise ValueError("Attribute {} does not exist.".format(item))

        cls.__init__ = __init__
        if not self._do_not_overwrite_setter:
            cls.__setattr__ = __setattr__
        cls.get_monitorable_state = get_monitorable_state
        cls.get_placeholder_value = get_placeholder_value
        cls.subscribe_attribute = subscribe_attribute
        cls.notify_virtual_change = _notify_placeholder_change
        cls.attribute_futures = defaultdict(lambda: defaultdict(list))

        return cls



"""Hold coil ejector."""
from mpf.devices.ball_device.ball_device_ejector import BallDeviceEjector


class HoldCoilEjector(BallDeviceEjector):

    """Hold balls by enabling and releases by disabling a coil."""

    __slots__ = ["hold_release_in_progress"]

    def eject_all_balls(self):
        """Eject all balls."""
        raise NotImplementedError()

    def __init__(self, config, ball_device, machine):
        """Initialise hold coil ejector."""
        for option in ["hold_switches", "hold_coil", "hold_coil_release_time", "hold_events"]:
            if option not in config and option in ball_device.config:
                config[option] = ball_device.config[option]
        super().__init__(config, ball_device, machine)
        self.hold_release_in_progress = False

        self.config = self.machine.config_validator.validate_config("ball_device_ejector_hold", self.config)

        # handle hold_coil activation when a ball hits a switch
        for switch in self.config['hold_switches']:
            if '{}_active'.format(self.ball_device.config['captures_from'].name) in switch.tags:
                self.ball_device.raise_config_error(
                    "Ball device '{}' uses switch '{}' which has a "
                    "'{}_active' tag. This is handled internally by the device. Remove the "
                    "redundant '{}_active' tag from that switch.".format(
                        self.ball_device.name, switch.name, self.ball_device.config['captures_from'].name,
                        self.ball_device.config['captures_from'].name), 13)

            self.ball_device.machine.switch_controller.add_switch_handler_obj(
                switch=switch, state=1,
                ms=0,
                callback=self.hold)

        for event in self.config["hold_events"]:
            self.machine.events.add_handler(event, self.hold)

    async def eject_one_ball(self, is_jammed, eject_try, balls_in_device):
        """Eject one ball by disabling hold coil."""
        del balls_in_device
        # TODO: wait for some time to allow balls to settle for
        #       both entrance and after a release
        self._disable_hold_coil()
        self.hold_release_in_progress = True

        # allow timed release of single balls and reenable coil after
        # release. Disable coil when device is empty
        self.ball_device.delay.add(name='hold_coil_release',
                                   ms=self.config['hold_coil_release_time'],
                                   callback=self._hold_release_done)
        # TODO: support ejecting a single ball by checking the ball_counter

    async def reorder_balls(self):
        """Do nothing."""
        # TODO: disable coil for a short period

    def _disable_hold_coil(self):
        self.config['hold_coil'].disable()
        self.ball_device.debug_log("Disabling hold coil. New "
                                   "balls: %s.", self.ball_device.balls)

    def hold(self, **kwargs):
        """Event handler for hold event."""
        del kwargs
        # do not enable coil when we are ejecting
        if self.hold_release_in_progress:
            return

        self._enable_hold_coil()

    def _enable_hold_coil(self):
        self.config['hold_coil'].enable()
        self.ball_device.debug_log("Enabling hold coil. New "
                                   "balls: %s.", self.ball_device.balls)

    def _hold_release_done(self):
        self.hold_release_in_progress = False
        self.ball_device.log.debug("No more balls. Hold coil will stay disabled.")

        # reenable hold coil if there are balls left
        if self.ball_device.balls > 0:
            self._enable_hold_coil()

    def ball_search(self, phase, iteration):
        """Run ball search."""
        self.config['hold_coil'].pulse()
        return True



from mpf.core.custom_code import CustomCode


class TestCustomCode(CustomCode):
    def on_load(self):
        self.log.debug("Loaded!")

        self.machine.events.add_handler('test_event', self._update)

    def _update(self, **kwargs):
        del kwargs
        self.machine.events.post("test_response")



"""Contains the DataManager base class."""

import copy
import os
import errno
import threading
import time
import _thread

from mpf.core.file_manager import FileManager
from mpf.core.mpf_controller import MpfController


class DataManager(MpfController):

    """Handles key value data loading and saving for the machine."""

    config_name = "data_manager"

    __slots_ = ["name", "min_wait_secs", "filename", "data", "_dirty"]

    def __init__(self, machine, name, min_wait_secs=1):
        """Initialise data manger.

        The DataManager is responsible for reading and writing data to/from a
        file on disk.

        Args:
        ----
            machine: The main MachineController instance.
            name: A string name that represents what this DataManager instance
                is for. This name is used to lookup the configuration option
                in the machine config in the mpf:paths:<name> location. That's
                how you specify the file name this DataManager will use.
            min_wait_secs: Minimal seconds to wait between two writes.
        """
        super().__init__(machine)
        self.name = name
        self.min_wait_secs = min_wait_secs
        config_path = self.machine.config['mpf']['paths'][name]
        if config_path is False:
            self.filename = False
        elif isinstance(config_path, str) and config_path.startswith("/"):
            self.filename = config_path
        elif isinstance(config_path, str):
            self.filename = os.path.join(self.machine.machine_path,
                                         self.machine.config['mpf']['paths'][name])
        else:
            raise AssertionError("Invalid path {} for {}".format(config_path, name))

        self.data = dict()
        self._dirty = threading.Event()

        if self.filename:
            self._setup_file()

            _thread.start_new_thread(self._writing_thread, ())

    def _setup_file(self):
        self._make_sure_path_exists(os.path.dirname(self.filename))

        self._load()

    @classmethod
    def _make_sure_path_exists(cls, path):
        try:
            os.makedirs(path)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise

    def _load(self):
        self.debug_log("Loading %s from %s", self.name, self.filename)
        if os.path.isfile(self.filename):
            self.data = FileManager.load(self.filename, halt_on_error=False)

        else:
            self.debug_log("Didn't find the %s file. No prob. We'll create "
                           "it when we save.", self.name)

        if not self.data:
            self.data = {}

    def get_data(self, section=None):
        """Return the value of this DataManager's data.

        Args:
        ----
            section: Optional string name of a section (dictionary key) for the
                data you want returned. Default is None which returns the
                entire dictionary.
        """
        if not section:
            data = copy.copy(self.data)
        else:

            try:
                data = copy.copy(self.data[section])
            except (KeyError, TypeError):
                data = dict()

        if isinstance(data, dict):
            return data

        return dict()

    def _trigger_save(self):
        """Trigger a write of this DataManager's data to the disk."""
        self.debug_log("Will write %s to disk", self.name)
        self._dirty.set()

    def save_all(self, data):
        """Update all data."""
        self.data = data
        self._trigger_save()

    def _writing_thread(self):  # pragma: no cover
        # prevent early writes at start-up
        time.sleep(self.min_wait_secs)
        while not self.machine.thread_stopper.is_set():
            if not self._dirty.wait(1):
                continue
            self._dirty.clear()

            data = copy.deepcopy(self.data)
            self.debug_log("Writing %s to: %s", self.name, self.filename)
            # save data
            FileManager.save(self.filename, data)
            # prevent too many writes
            time.sleep(self.min_wait_secs)

        # if dirty write data one last time during shutdown
        if self._dirty.is_set():
            FileManager.save(self.filename, data)



# pylint: disable-msg=too-many-lines
"""Stern Spike Platform."""
import asyncio

import random
from typing import Optional, Union

from mpf.platforms.interfaces.light_platform_interface import LightPlatformSoftwareFade
from mpf.platforms.interfaces.stepper_platform_interface import StepperPlatformInterface
from mpf.core.platform_batch_light_system import PlatformBatchLight, PlatformBatchLightSystem
from mpf.platforms.interfaces.dmd_platform import DmdPlatformInterface
from mpf.platforms.interfaces.driver_platform_interface import DriverPlatformInterface, PulseSettings, HoldSettings
from mpf.platforms.interfaces.switch_platform_interface import SwitchPlatformInterface
from mpf.platforms.spike.spike_defines import SpikeNodebus
from mpf.core.platform import SwitchPlatform, DriverPlatform, LightsPlatform, SwitchSettings, DriverSettings, \
    DriverConfig, SwitchConfig, DmdPlatform, StepperPlatform, RepulseSettings

from mpf.core.utility_functions import Util


class SpikeSwitch(SwitchPlatformInterface):

    """A switch on a Stern Spike node board."""

    __slots__ = ["node", "index", "platform", "platform_config", "_last_debounce"]

    def __init__(self, config, number, platform, platform_config):
        """Initialise switch."""
        super().__init__(config, number)
        self.node, self.index = self.number.split("-")
        self.node = int(self.node)
        self.index = int(self.index)
        self.platform = platform
        self.platform_config = platform_config
        self._last_debounce = None

    def configure_debounce_and_recycle(self, enable_debounce, invert_switch, enable_recycle):
        """Configure debounce time."""
        if self.platform.node_firmware_version[self.node] < 0x3100:
            return

        # the logic here is a bit tricky because we use debounce to implement recycle as well
        # if recycle is enabled we will always debounce the switch closing (or opening if inverted)
        # this kind of emulates recycling

        if enable_debounce and (not invert_switch or not enable_recycle):
            if self.platform_config["debounce_open"]:
                debounce_open_ms = self.platform_config["debounce_open"]
            else:
                debounce_open_ms = self.platform.config["default_debounce_open"]
        else:
            debounce_open_ms = 2

        if enable_debounce and (invert_switch or not enable_recycle):
            if self.platform_config["debounce_close"]:
                debounce_close_ms = self.platform_config["debounce_close"]
            else:
                debounce_close_ms = self.platform.config["default_debounce_close"]
        else:
            debounce_close_ms = 2

        new_debounce = (debounce_open_ms, debounce_close_ms)
        if new_debounce == self._last_debounce:
            return

        self._last_debounce = new_debounce

        debounce_open_value = int(debounce_open_ms * self.platform.ticks_per_sec[self.node] / 1000)
        debounce_closed_value = int(debounce_close_ms * self.platform.ticks_per_sec[self.node] / 1000)
        if debounce_open_value > 0xfe:
            debounce_open_value = 0xfe
        if debounce_closed_value > 0xfe:
            debounce_closed_value = 0xfe

        self.platform.log.debug("Set debounce open: %s close: %s (about %sms and %sms) for switch %s",
                                debounce_open_value, debounce_closed_value, debounce_open_value,
                                debounce_closed_value, self.number)
        self.platform.send_cmd_async(self.node, SpikeNodebus.ConfigInput, bytearray([self.index,
                                                                                     debounce_open_value,
                                                                                     debounce_closed_value]))

    def get_board_name(self):
        """Return name for service mode."""
        return "Spike Node {}".format(self.node)


class SpikeBacklight(LightPlatformSoftwareFade):

    """The backlight on the CPU node."""

    __slots__ = ["platform"]

    def __init__(self, number, platform, loop, fade_interval_ms):
        """Initialise backlight."""
        super().__init__(number, loop, fade_interval_ms)
        self.platform = platform        # type: SpikePlatform

    def set_brightness(self, brightness: float):
        """Set brightness."""
        # we use the Spike 2 command. it will be intercepted in the mpf-spike bridge for Spike 1
        brightness_u16 = int(brightness * 65535)
        self.platform.send_cmd_raw_async([SpikeNodebus.SetBackboxLight, 2,
                                          brightness_u16 >> 8, brightness_u16 & 0xFF,
                                          0])

    def get_board_name(self):
        """Return name for service mode."""
        return "Spike Node 0"

    def is_successor_of(self, other):
        """There can only be one backlight."""
        raise AssertionError("Not possible for backlight.")

    def get_successor_number(self):
        """There can only be one backlight."""
        raise AssertionError("Not possible for backlight.")

    def __lt__(self, other):
        """There can only be one backlight."""
        return self.number < other.number


class SpikeLight(PlatformBatchLight):

    """A light on a Stern Spike node board."""

    __slots__ = ["node", "index", "platform", "_max_fade"]

    def __init__(self, number, platform, light_system):
        """Initialise light."""
        super().__init__(number, light_system)
        node, index = number.split("-")
        self.node = int(node)
        self.index = int(index)
        self.platform = platform
        self._max_fade = None

    def get_max_fade_ms(self):
        """Return max fade ms."""
        if self.node == 0:
            return 0
        if self._max_fade is None:
            # calculate max fade time. 255 is reserved so 254 is the max value
            self._max_fade = int(254 * 1000 / self.platform.ticks_per_sec[self.node])
        return self._max_fade

    def get_board_name(self):
        """Return name for service mode."""
        return "Spike Node {}".format(self.node)

    def is_successor_of(self, other):
        """Return true if the other light has the previous index and is on the same node."""
        return self.node == other.node and self.index == other.index + 1

    def get_successor_number(self):
        """Return next index on node."""
        return self.node, self.index + 1

    def __lt__(self, other):
        """Order lights by their order on the hardware."""
        return (self.node, self.index) < (other.node, other.index)


class SpikeDMD(DmdPlatformInterface):

    """The DMD on the SPIKE system."""

    __slots__ = ["platform", "data", "new_frame_event", "dmd_task"]

    def __init__(self, platform):
        """Initialise DMD."""
        self.platform = platform
        self.data = None
        self.new_frame_event = asyncio.Event()
        self.dmd_task = platform.machine.clock.loop.create_task(self._dmd_send())
        self.dmd_task.add_done_callback(Util.raise_exceptions)

    def update(self, data: bytes):
        """Remember the last frame data."""
        self.data = data
        self.new_frame_event.set()

    async def _dmd_send(self):
        while True:
            await self.new_frame_event.wait()
            self.new_frame_event.clear()
            await self.send_update()
            # sleep at least 5ms
            await asyncio.sleep(.005)

    async def send_update(self):
        """Send update to platform."""
        if len(self.data) != 128 * 32:
            raise AssertionError("Invalid frame length for SPIKE. Should be 128*32 pixels.")
        frame1 = bytearray()
        frame2 = bytearray()
        frame3 = bytearray()
        frame4 = bytearray()
        # we build four frames for a 128*32 pixel display. one bit per pixel each = 512bytes
        for i in range(512):
            pixel1 = 0
            pixel2 = 0
            pixel3 = 0
            pixel4 = 0
            for p in range(8):
                pixel_data = self.data[i * 8 + p]
                pixel1 += 1 if pixel_data & 0x01 else 0
                pixel2 += 1 if pixel_data & 0x02 else 0
                pixel3 += 1 if pixel_data & 0x04 else 0
                pixel4 += 1 if pixel_data & 0x08 else 0
                pixel1 *= 2
                pixel2 *= 2
                pixel3 *= 2
                pixel4 *= 2

            frame1.append(int(pixel1 / 2))
            frame2.append(int(pixel2 / 2))
            frame3.append(int(pixel3 / 2))
            frame4.append(int(pixel4 / 2))
        await self.platform.send_cmd_raw(bytes([0x80, 0x00, 0x90]) + bytes(frame1) + bytes(frame2) +
                                         bytes(frame3) + bytes(frame4))

    def set_brightness(self, brightness: float):
        """Set brightness of the DMD."""
        # we do not yet know how that works in SPIKE


class SpikeDriver(DriverPlatformInterface):

    """A driver on a Stern Spike node board."""

    __slots__ = ["platform", "node", "index", "_enable_task"]

    def __init__(self, config, number, platform):
        """Initialise driver on Stern Spike."""
        super().__init__(config, number)
        self.platform = platform
        self.node, self.index = number.split("-")
        self.node = int(self.node)
        self.index = int(self.index)
        self._enable_task = None

    def trigger(self, power1, duration1, power2, duration2):
        """Pulse the driver."""
        msg = bytearray([
            self.index,
            power1,
            duration1 & 0xFF,
            1 if duration1 & 0x100 else 0,
            power2,
            duration2 & 0xFF,
            1 if duration2 & 0x100 else 0,
            0,
            0
        ])
        self.platform.send_cmd_async(self.node, SpikeNodebus.CoilFireRelease, msg)

    def enable(self, pulse_settings: PulseSettings, hold_settings: HoldSettings):
        """Pulse and enable coil."""
        if self._enable_task:
            return

        # initial pulse
        power1 = int(pulse_settings.power * 255)
        duration1 = int(pulse_settings.duration * self.platform.ticks_per_sec[self.node] / 1000)

        if duration1 > 0x1FF:
            raise AssertionError("Initial pulse too long.")

        # then enable hold
        power2 = int(hold_settings.power * 255)
        duration2 = 0x1FF

        self.trigger(power1, duration1, power2, duration2)

        self._enable_task = self.platform.machine.clock.loop.create_task(self._enable(power2))
        self._enable_task.add_done_callback(Util.raise_exceptions)

    async def _enable(self, power):
        while True:
            await asyncio.sleep(.25)
            self.trigger(power, 0x1FF, power, 0x1FF)

    def pulse(self, pulse_settings: PulseSettings):
        """Pulse coil for a certain time."""
        power1 = power2 = int(pulse_settings.power * 255)
        duration1 = int(pulse_settings.duration * self.platform.ticks_per_sec[self.node] / 1000)
        duration2 = duration1 - 0x1FF if duration1 > 0x1FF else 0
        if duration2 > 0x1FF:
            raise AssertionError("Pulse ms too long.")

        self.trigger(power1, duration1, power2, duration2)

    def disable(self):
        """Disable coil."""
        # cancel enable task
        if self._enable_task:
            self._enable_task.cancel()
            self._enable_task = None

        # disable coil
        self.trigger(0, 0, 0, 0)

    def get_board_name(self):
        """Return name for service mode."""
        return "Spike Node {}".format(self.node)


class SpikeStepper(StepperPlatformInterface):

    """A stepper in Spike."""

    __slots__ = ["number", "node", "stepper_id", "config", "platform", "_position", "light_index"]

    def __init__(self, number, config, platform):
        """Initialise stepper."""
        self.number = number
        node, index = number.split("-", 2)
        self.node = int(node)
        self.stepper_id = int(index)
        self.config = config
        self.platform = platform    # type: SpikePlatform
        self._position = 0
        light_node, light_index = self.config['light_number'].split("-", 2)
        if light_node != node:
            self.platform.raise_config_error("Light and Stepper have to be on the same node."
                                             " Light: {} Stepper: {}".format(number, self.config['light_number']), 1)
        self.light_index = int(light_index)
        self._configure()

    def home(self, direction):
        """Unsupported."""
        raise AssertionError("Use a switch.")

    def set_home_position(self):
        """Set position to home."""
        self._mark_as_home()

    def _configure(self):
        """Configure stepper in spike."""
        unknown1 = 200
        unknown2 = 0
        unknown3 = 0xff
        self.platform.send_cmd_async(self.node, SpikeNodebus.StepperConfig, [
            self.stepper_id,
            unknown1,                   # unclear what it does
            unknown2 & 0xff,            # unclear what it does
            (unknown2 >> 8) & 0xff,
            self.light_index + 0x40,
            unknown3                    # unclear what it does
        ])

    def _mark_as_home(self):
        """Tell the node that the stepper is home."""
        self.platform.send_cmd_async(self.node, SpikeNodebus.StepperHome, [
            self.stepper_id
        ])

    async def _get_stepper_info(self):
        """Get info about the stepper."""
        '''
        position (u16), unknown (u16), flags (u8)
        flags:
            0 - unknown
            1 - is_active
            2 - unknown
            3 - unknown
            4 - unknown
            5 - unknown
            6 - unknown
            7 - unknown

        Typical response: 17 00 18 00 01 d0 00
        '''
        result = await self.platform.send_cmd_and_wait_for_response(
            self.node, SpikeNodebus.StepperInfo + self.stepper_id, [], 7)
        if result is None:
            self.platform.warning_log("Failed to read stepper %s position", self.number)
            return None
        return {
            "position": result[0] + (result[1] << 8),
            "is_active": bool((result[4] >> 1) & 1)
        }

    def _move_to_absolute_position(self, position, speed):
        """Move the stepper to a certain position."""
        self.platform.send_cmd_async(self.node, SpikeNodebus.StepperSet, [
            self.stepper_id,
            position & 0xff,
            (position >> 8) & 0xff,
            speed
        ])

    async def wait_for_move_completed(self):
        """Wait until move completed."""
        while not await self.is_move_complete():
            await asyncio.sleep(1 / self.config['poll_ms'])

    async def is_move_complete(self) -> bool:
        """Return true if move is complete."""
        info = await self._get_stepper_info()
        if info is None:
            # read did fail
            return False
        return info['position'] == self._position

    def move_rel_pos(self, position):
        """Move relative to current position."""
        self._position += int(position)
        self._move_to_absolute_position(self._position, self.config['speed'])

    def move_vel_mode(self, velocity):
        """Move stepper in a direction."""
        # this is not really supported and we could do better here
        # also we do not know how to move to the other direciton
        if velocity < 0:
            raise AssertionError("We do not know how to do this in Spike. Let us know if you need it.")
        self._move_to_absolute_position(200, self.config['homing_speed'])

    def stop(self):
        """Stop the stepper."""
        # reconfiguring seems to be used to stop the stepper
        self._configure()


# pylint: disable-msg=too-many-instance-attributes
class SpikePlatform(SwitchPlatform, LightsPlatform, DriverPlatform, DmdPlatform, StepperPlatform):

    """Stern Spike Platform."""

    __slots__ = ["_writer", "_reader", "_inputs", "config", "_poll_task", "_sender_task", "_send_key_task", "dmd",
                 "_nodes", "_bus_read", "_bus_write", "_cmd_queue", "ticks_per_sec", "_light_system",
                 "node_firmware_version", "_query_nodes_task"]

    def __init__(self, machine):
        """Initialise spike hardware platform."""
        super().__init__(machine)
        self._writer = None
        self._reader = None
        self._inputs = {}
        self._poll_task = None
        self._sender_task = None
        self._send_key_task = None
        self._query_nodes_task = None
        self.dmd = None
        self.features['has_lights'] = True

        self._nodes = None
        self._bus_read = asyncio.Lock()
        self._bus_write = asyncio.Lock()
        self._cmd_queue = asyncio.Queue()

        self._light_system = None

        self.ticks_per_sec = {
            0: 1
        }
        self.node_firmware_version = {
            0: 0x000000
        }

        self.config = self.machine.config_validator.validate_config("spike", self.machine.config['spike'])
        self._configure_device_logging_and_debug("Spike", self.config)

    @classmethod
    def get_switch_config_section(cls):
        """Return switch config section."""
        return "spike_switch_overwrites"

    async def _send_multiple_light_update(self, sequential_brightness_list):
        common_fade_ms = sequential_brightness_list[0][2]
        if common_fade_ms < 0:
            common_fade_ms = 0
        fade_time = int(common_fade_ms * self.ticks_per_sec[sequential_brightness_list[0][0].node] / 1000)

        data = bytearray([fade_time])
        for _, brightness, _ in sequential_brightness_list:
            data.append(int(255 * brightness))

        self.send_cmd_async(sequential_brightness_list[0][0].node,
                            SpikeNodebus.SetLed + sequential_brightness_list[0][0].index, data)

    # pylint: disable-msg=too-many-arguments
    def _write_rule(self, node, enable_switch_index, disable_switch_index, coil_index, pulse_settings: PulseSettings,
                    hold_settings: Optional[HoldSettings], param1, param2, param3, hold_param1=0):
        """Write a hardware rule to Stern Spike.

        [14] is used in CoilReflex. Might be debounce? (NODEBUS_CoilConfig param_3[1] (ushort))
            Slings = 0xf5
            Pops = 0x14
            Flippers = 0

        We do not yet understand param1, param2 and param3:
        param1 == 2 -> second switch (eos switch)
        param2 == 1 -> ??
        param2 == 6 -> ??
        param3 == 5 -> allow enable
        """
        pulse_value = int(pulse_settings.duration * self.ticks_per_sec[node] / 1000)

        if self.node_firmware_version[node] >= 0x2800:
            # length: 0x24
            self.send_cmd_async(node, SpikeNodebus.CoilSetReflex, bytearray(
                [coil_index,                                                # coil [3]
                 int(pulse_settings.power * 255),                           # pulse power (coil1)[4]
                 pulse_value & 0xFF,                                        # pulse time lower (coil1) [5]
                 (pulse_value & 0xFF00) >> 8,                               # pulse time upper (coil2) [6]
                 int(hold_settings.power * 255) if hold_settings else 0,    # hold power (coil1) [7]
                 hold_param1,                                               # some time lower (probably pulse coil2) [8]
                 0x00,                                                      # some time upper (probably pulse coil2) [9]
                 0x00,                                                      # some time lower (unknown) [10]
                 0x00,                                                      # some time upper (unknown) [11]
                 0,                                                         # a unknown time lower (1) [12]
                 0,                                                         # a unknown time upper (1) [13]
                 0,                                                         # a unknown time lower (2) [14]
                 0,                                                         # a unknown time upper (2) [15]
                 0,                                                         # a unknown time lower (3) [16]
                 0,                                                         # a unknown time upper (3) [17]
                 0,                                                         # a unknown time lower (4) [18]
                 0,                                                         # a unknown time upper (4) [19]
                 0,                                                         # a unknown time lower (5) [20]
                 0,                                                         # a unknown time upper (5) [21]
                 0x40 ^ enable_switch_index,                                # enable switch [22]
                 0,                                                         # another switch? [23]
                 0,                                                         # another switch? [24]
                 0x40 ^ disable_switch_index if disable_switch_index is not None else 0,  # disable switch [25]
                 0,                                                         # another switch? [26]
                 0,                                                         # another switch? [27]
                 0,                                                         # another switch? [28]
                 0,                                                         # another switch? [29]
                 0,                                                         # param5 [30]
                 0,                                                         # param5 [31]
                 0,                                                         # param5 [32]
                 0,                                                         # param5 [33]
                 param1,                                                    # some time (param6) [34]
                 param2,                                                    # param7 [35]
                 param3                                                     # param8 [36]
                 ]))
        else:
            # length: 0x19
            self.send_cmd_async(node, SpikeNodebus.CoilSetReflex, bytearray(
                [coil_index,                                                # coil [3]
                 int(pulse_settings.power * 255),                           # pulse power [4]
                 pulse_value & 0xFF,                                        # pulse time lower [5]
                 (pulse_value & 0xFF00) >> 8,                               # pulse time upper [6]
                 int(hold_settings.power * 255) if hold_settings else 0,    # hold power [7]
                 0x00,                                                      # some time lower (probably hold) [8]
                 0x00,                                                      # some time upper (probably hold) [9]
                 0x00,                                                      # some time lower (unknown) [10]
                 0x00,                                                      # some time upper (unknown) [11]
                 0,                                                         # a unknown time lower (1) [12]
                 0,                                                         # a unknown time upper (1) [13]
                 0,                                                         # a unknown time lower (2) [14]
                 0,                                                         # a unknown time upper (2) [15]
                 0,                                                         # a unknown time lower (3) [16]
                 0,                                                         # a unknown time upper (3) [17]
                 0,                                                         # a unknown time lower (4) [18]
                 0,                                                         # a unknown time upper (4) [19]
                 0x40 ^ enable_switch_index,                                # enable switch [20]
                 0x40 ^ disable_switch_index if disable_switch_index is not None else 0,    # disable switch [21]
                 0,                                                         # another switch? [22]
                 param1,                                                    # param5 [23]
                 param2,                                                    # some time (param6) [24]
                 param3                                                     # param7 [25]
                 ]))

    @staticmethod
    def _check_coil_switch_combination(switch, coil):
        if switch.hw_switch.node != coil.hw_driver.node:
            raise AssertionError("Coil {} and Switch {} need to be on the same node to write a rule".format(
                coil, switch
            ))

    def set_pulse_on_hit_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit rule on driver.

        This is mostly used for popbumpers. Example from WWE:
        Type: 8 Cmd: 65 Node: 9 Msg: 0x00 0xa6 0x28 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x14 0x00 0x00 0x00 0x38
        0x00 0x40 0x00 0x00 0x00 0x00 0x00 Len: 25
        """
        self._check_coil_switch_combination(enable_switch, coil)
        enable_switch.hw_switch.configure_debounce_and_recycle(enable_switch.debounce, enable_switch.invert,
                                                               coil.recycle)
        self._write_rule(coil.hw_driver.node, enable_switch.hw_switch.index ^ (enable_switch.invert * 0x40),
                         None, coil.hw_driver.index,
                         coil.pulse_settings, None, 0, 0, 0)

    def set_pulse_on_hit_and_enable_and_release_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit and enable and release rule on driver.

        Used for single coil flippers. Examples from WWE:
        Dual-wound flipper hold coil:
        Type: 8 Cmd: 65 Node: 8 Msg: 0x02 0xff 0x46 0x01 0xff 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x3a
        0x00 0x42 0x40 0x00 0x00 0x01 0x00  Len: 25

        Ring Slings (different flags):
        Type: 8 Cmd: 65 Node: 10 Msg: 0x00 0xff 0x19 0x00 0x14 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x80
        0x00 0x4a 0x40 0x00 0x00 0x06 0x05  Len: 25
        """
        self._check_coil_switch_combination(enable_switch, coil)
        enable_switch.hw_switch.configure_debounce_and_recycle(enable_switch.debounce, enable_switch.invert,
                                                               coil.recycle)
        self._write_rule(coil.hw_driver.node, enable_switch.hw_switch.index ^ (enable_switch.invert * 0x40),
                         None, coil.hw_driver.index,
                         coil.pulse_settings, coil.hold_settings, 0, 6, 5)

    def set_pulse_on_hit_and_release_and_disable_rule(self, enable_switch: SwitchSettings, eos_switch: SwitchSettings,
                                                      coil: DriverSettings,
                                                      repulse_settings: Optional[RepulseSettings]):
        """Set pulse on hit and release rule to driver.

        Used for high-power coil on dual-wound flippers. Example from WWE:
        Type: 8 Cmd: 65 Node: 8 Msg: 0x00 0xff 0x33 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00
        0x00 0x42 0x40 0x00 0x02 0x06 0x00  Len: 25
        """
        self._check_coil_switch_combination(enable_switch, coil)
        self._check_coil_switch_combination(eos_switch, coil)

        if self.node_firmware_version[coil.hw_driver.node] >= 0x2800:
            flag1 = 5
            hold_param1 = 0x28
        else:
            flag1 = 2
            hold_param1 = 0

        enable_switch.hw_switch.configure_debounce_and_recycle(enable_switch.debounce, enable_switch.invert,
                                                               coil.recycle)
        eos_switch.hw_switch.configure_debounce_and_recycle(eos_switch.debounce, eos_switch.invert,
                                                            coil.recycle)

        # this might have to be the same flags as set_pulse_on_hit_and_enable_and_release_and_disable_rule
        # it might also just depend on Spike System 1 vs 2
        self._write_rule(coil.hw_driver.node, enable_switch.hw_switch.index ^ (enable_switch.invert * 0x40),
                         eos_switch.hw_switch.index ^ (eos_switch.invert * 0x40),
                         coil.hw_driver.index, coil.pulse_settings, coil.hold_settings, flag1, 6, 0,
                         hold_param1=hold_param1)

    def set_pulse_on_hit_and_enable_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                                 eos_switch: SwitchSettings, coil: DriverSettings,
                                                                 repulse_settings: Optional[RepulseSettings]):
        """Set pulse on hit and release rule to driver.

        Used for high-power coil on single-wound flippers. Example from GoT:
        Node: 8 Command CoilSetReflex (0x41) Params: 0x00 0xff 0x33 0x00 0x1e 0x28 0x00 0x00 0x00 0x00 0x00 0x00 0x00
        0x00 0x00 0x00 0x00 0x00 0x00 0x42 0x00 0x00 0x40 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x05 0x06 0x00 Len: 36
        """
        self._check_coil_switch_combination(enable_switch, coil)
        self._check_coil_switch_combination(eos_switch, coil)

        if self.node_firmware_version[coil.hw_driver.node] >= 0x2800:
            flag1 = 5
            hold_param1 = 0x28
        else:
            flag1 = 2
            hold_param1 = 0
        enable_switch.hw_switch.configure_debounce_and_recycle(enable_switch.debounce, enable_switch.invert,
                                                               coil.recycle)
        eos_switch.hw_switch.configure_debounce_and_recycle(eos_switch.debounce, eos_switch.invert,
                                                            coil.recycle)
        self._write_rule(coil.hw_driver.node, enable_switch.hw_switch.index ^ (enable_switch.invert * 0x40),
                         eos_switch.hw_switch.index ^ (eos_switch.invert * 0x40),
                         coil.hw_driver.index, coil.pulse_settings, coil.hold_settings, flag1, 6, 0,
                         hold_param1=hold_param1)

    def set_pulse_on_hit_and_release_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit and release rule to driver.

        I believe that param2 == 1 means that it will cancel the pulse when the switch is released.

        Used for high-power coils on dual-wound flippers. Example from WWE:
        Type: 8 Cmd: 65 Node: 8 Msg: 0x03 0xff 0x46 0x01 0xff 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00
        0x00 0x43 0x40 0x00 0x00 0x01 0x00  Len: 25

        """
        self._check_coil_switch_combination(enable_switch, coil)
        enable_switch.hw_switch.configure_debounce_and_recycle(enable_switch.debounce, enable_switch.invert,
                                                               coil.recycle)
        self._write_rule(coil.hw_driver.node, enable_switch.hw_switch.index ^ (enable_switch.invert * 0x40),
                         None, coil.hw_driver.index,
                         coil.pulse_settings, None, 0, 1, 0)

    def clear_hw_rule(self, switch, coil):
        """Disable hardware rule for this coil."""
        del switch
        if self.node_firmware_version[coil.hw_driver.node] >= 0x2800:
            self.send_cmd_async(coil.hw_driver.node, SpikeNodebus.CoilSetReflex, bytearray(
                [coil.hw_driver.index] + [0x00] * 33))
        else:
            self.send_cmd_async(coil.hw_driver.node, SpikeNodebus.CoilSetReflex, bytearray(
                [coil.hw_driver.index, 0, 0, 0, 0, 0x00, 0x00, 0x00, 0x00,
                 0, 0, 0, 0, 0, 0, 0, 0,
                 0, 0, 0,
                 0, 0, 0]))

    def configure_driver(self, config: DriverConfig, number: str, platform_settings: dict):
        """Configure a driver on Stern Spike."""
        del platform_settings

        try:
            node, _ = number.split("-")
        except IndexError:
            return self.raise_config_error("Coil number has to have the syntax node-index but is:"
                                           " {} for driver {}".format(number, config.name),
                                           4)

        if int(node) not in self._nodes:
            self.raise_config_error("Node {} not known for coil {}".format(node, number), 5)

        return SpikeDriver(config, number, self)

    def parse_light_number_to_channels(self, number: str, subtype: str):
        """Return a single light."""
        return [
            {
                "number": number,
            }
        ]

    def configure_light(self, number, subtype, config, platform_settings) -> Union[SpikeLight, SpikeBacklight]:
        """Configure a light on Stern Spike."""
        del platform_settings, subtype
        if number == "0-0":
            return SpikeBacklight(number, self, self.machine.clock.loop, 3)

        try:
            node, index = number.split("-")
        except IndexError:
            return self.raise_config_error("Light number has to have the syntax node-index but is:"
                                           " {} for light {}".format(number, config.name),
                                           7)

        if int(node) in self.config['node_config'] and \
                self.config['node_config'][int(node)]['num_leds'] is not None and \
                int(index) > self.config['node_config'][int(node)]['num_leds']:
            return self.raise_config_error("Light {} is larger than the maximum configured led {} for "
                                           "node {}".format(number,
                                                            self.config['node_config'][int(node)]['num_leds'],
                                                            node), 8)

        # TODO: validate that light number is not used in a stepper and a light
        return SpikeLight(number, self, self._light_system)

    def configure_switch(self, number: str, config: SwitchConfig, platform_config: dict):
        """Configure switch on Stern Spike."""
        try:
            node, index = number.split("-")
        except IndexError:
            return self.raise_config_error("Switch number has to have the syntax node-index but is:"
                                           " {} for switch {}".format(number, config.name),
                                           2)

        if int(node) not in self._nodes:
            self.raise_config_error("Node {} not known for switch {}".format(node, number), 3)

        if int(node) in self.config['node_config'] and \
                self.config['node_config'][int(node)]['num_inputs'] is not None and \
                int(index) > self.config['node_config'][int(node)]['num_inputs']:
            return self.raise_config_error("Switch {} is larger than the maximum configured input {} for "
                                           "node {}".format(number,
                                                            self.config['node_config'][int(node)]['num_inputs'],
                                                            node), 6)

        switch = SpikeSwitch(config, number, self, platform_config)
        switch.configure_debounce_and_recycle(config.debounce, config.invert, False)
        return switch

    async def get_hw_switch_states(self):
        """Return current switch states."""
        hw_states = dict()
        for node in self._nodes:
            curr_bit = 1
            for index in range(0, 64):
                hw_states[str(node) + '-' + str(index)] = (curr_bit & self._inputs[node]) == 0
                curr_bit <<= 1
        return hw_states

    async def initialize(self):
        """Initialise platform."""
        port = self.config['port']
        baud = self.config['baud']
        flow_control = self.config['flow_control']
        self.debug = self.config['debug']
        self._nodes = self.config['nodes']

        if 0 not in self._nodes:
            raise AssertionError("Please include CPU node 0 in nodes for Spike.")

        await self._connect_to_hardware(port, baud, flow_control=flow_control)

        self._poll_task = self.machine.clock.loop.create_task(self._poll())
        self._poll_task.add_done_callback(Util.raise_exceptions)

        self._sender_task = self.machine.clock.loop.create_task(self._sender())
        self._sender_task.add_done_callback(Util.raise_exceptions)

        if self.config['use_send_key']:
            self._send_key_task = self.machine.clock.loop.create_task(self._send_key())
            self._send_key_task.add_done_callback(Util.raise_exceptions)

        if self.config['periodically_query_nodes']:
            self._query_nodes_task = self.machine.clock.loop.create_task(self._query_status_and_coil_current())
            self._query_nodes_task.add_done_callback(Util.raise_exceptions)

        self._light_system = PlatformBatchLightSystem(self.machine.clock, self._send_multiple_light_update,
                                                      self.machine.config['mpf']['default_light_hw_update_hz'],
                                                      self.config['max_led_batch_size'])
        self._light_system.start()

    async def _connect_to_hardware(self, port, baud, *, flow_control=False, xonxoff=False):
        self.info_log("Connecting to %s at %sbps", port, baud)

        connector = self.machine.clock.open_serial_connection(
            url=port, baudrate=baud, rtscts=flow_control, xonxoff=xonxoff)
        self._reader, self._writer = await connector
        self._writer.transport.set_write_buffer_limits(2048, 1024)

        await self._initialize()

    async def _update_switches(self, node):
        if node not in self._nodes:     # pragma: no cover
            self.warning_log("Cannot read node %s because it is not configured.", node)
            return False

        new_inputs_str = await self._read_inputs(node)
        if not new_inputs_str:      # pragma: no cover
            self.info_log("Node: %s did not return any inputs.", node)
            return True

        new_inputs = self._input_to_int(new_inputs_str)

        if self.debug:
            self.debug_log("Inputs node: %s State: %s Old: %s New: %s",
                           node, "".join(bin(b) + " " for b in new_inputs_str[0:8]), self._inputs[node], new_inputs)

        changes = self._inputs[node] ^ new_inputs
        if changes != 0:
            curr_bit = 1
            for index in range(0, 64):
                if (curr_bit & changes) != 0:
                    self.machine.switch_controller.process_switch_by_num(
                        state=(curr_bit & new_inputs) == 0,
                        num=str(node) + "-" + str(index),
                        platform=self)
                curr_bit <<= 1
        elif self.debug:    # pragma: no cover
            self.debug_log("Got input activity but inputs did not change.")

        self._inputs[node] = new_inputs

        return True

    async def _sender(self):
        while True:
            cmd, wait_ms = await self._cmd_queue.get()
            await self._send_cmd_without_response(cmd, wait_ms)

    async def _send_key(self):
        while True:
            for node in self._nodes:
                # do not send watchdog to cpu since it is a virtual node
                if node == 0:
                    continue

                # generate super secret "key"
                key = bytearray()
                for _ in range(16):
                    key.append(random.randint(0, 255))

                # send SendKey message
                await self.send_cmd_sync(node, SpikeNodebus.SendKey, key)

                # wait one second
                await asyncio.sleep(1.0)

    async def _query_status_and_coil_current(self):
        while True:
            for node in self._nodes:
                # do not query virtual node
                if node == 0:
                    continue

                self.debug_log("GetStatus and GetCoilCurrent on node %s", node)
                await self.send_cmd_and_wait_for_response(node, SpikeNodebus.GetStatus, bytearray(), 10)
                await self.send_cmd_and_wait_for_response(node, SpikeNodebus.GetCoilCurrent, bytearray([0]), 12)

                # wait before querying the next board
                await asyncio.sleep(.5)

    async def _poll(self):
        while True:
            async with self._bus_read:
                async with self._bus_write:
                    await self._send_raw(bytearray([0]))

                try:
                    result = await asyncio.wait_for(self._read_raw(1), 2)
                except asyncio.TimeoutError:    # pragma: no cover
                    self.warning_log("Spike watchdog expired.")
                    # clear buffer
                    # pylint: disable-msg=protected-access
                    self._reader._buffer = bytearray()
                    continue

            if not result:
                self.warning_log("Empty poll result. Spike desynced.")
                # give it a break of 50ms
                await asyncio.sleep(.05)
                # clear buffer
                # pylint: disable-msg=protected-access
                self._reader._buffer = bytearray()
                continue

            ready_node = result[0]

            if 0 < ready_node <= 0x0F or ready_node == 0xF0:
                # valid node ids
                if ready_node == 0xF0:
                    # virtual cpu node returns 0xF0 instead of 0 to make it distinguishable
                    ready_node = 0
                result = await self._update_switches(ready_node)
                if not result:
                    self.warning_log("Spike desynced during input.")
                    await asyncio.sleep(.05)
                    # clear buffer
                    # pylint: disable-msg=protected-access
                    self._reader._buffer = bytearray()
            elif ready_node > 0:    # pragma: no cover
                # invalid node ids
                self.warning_log("Spike desynced (invalid node %s).", ready_node)
                # give it a break of 50ms
                await asyncio.sleep(.05)
                # clear buffer
                # pylint: disable-msg=protected-access
                self._reader._buffer = bytearray()
            else:
                # sleep only if spike is idle
                await asyncio.sleep(1 / self.config['poll_hz'])

    def stop(self):
        """Stop hardware and close connections."""
        if self._light_system:
            self._light_system.stop()
        if self._poll_task:
            self._poll_task.cancel()
            try:
                self.machine.clock.loop.run_until_complete(self._poll_task)
            except asyncio.CancelledError:
                pass

        if self._sender_task:
            self._sender_task.cancel()
            try:
                self.machine.clock.loop.run_until_complete(self._sender_task)
            except asyncio.CancelledError:
                pass

        if self._send_key_task:
            self._send_key_task.cancel()
            try:
                self.machine.clock.loop.run_until_complete(self._send_key_task)
            except asyncio.CancelledError:
                pass

        if self._query_nodes_task:
            self._query_nodes_task.cancel()
            try:
                self.machine.clock.loop.run_until_complete(self._query_nodes_task)
            except asyncio.CancelledError:
                pass

        if self._writer:
            # shutdown the bridge
            self._writer.write(b'\xf5')
            # send ctrl+c to stop the mpf-spike-bridge
            self._writer.write(b'\x03reset\n')
            self._writer.close()

    async def _send_raw(self, data):
        if self.debug:
            self.debug_log("Sending: %s", "".join("%02x " % b for b in data))
        for start in range(0, len(data), 256):
            block = data[start:start + 256]
            self._writer.write(bytes(block))
        await self._writer.drain()

    async def _read_raw(self, msg_len: int) -> bytearray:
        if not msg_len:
            raise AssertionError("Cannot read 0 length")

        if self.debug:
            self.debug_log("Reading %s bytes", msg_len)

        data = await self._reader.readexactly(msg_len)

        if self.debug:
            self.debug_log("Data: %s", "".join("%02x " % b for b in data))

        return data

    @staticmethod
    def _checksum(cmd_str):
        checksum = 0
        for i in cmd_str:
            checksum += i
        return (256 - (checksum % 256)) % 256

    async def send_cmd_and_wait_for_response(self, node, cmd, data, response_len) -> Optional[bytearray]:
        """Send cmd and wait for response."""
        assert response_len > 0
        if node > 15:
            raise AssertionError("Node must be 0-15.")
        cmd_str = bytearray()
        cmd_str.append((8 << 4) + node)
        cmd_str.append(len(data) + 2)
        cmd_str.append(cmd)
        cmd_str.extend(data)
        cmd_str.append(self._checksum(cmd_str))
        cmd_str.append(response_len)
        async with self._bus_read:
            async with self._bus_write:
                await self._send_raw(cmd_str)
            try:
                response = await asyncio.wait_for(self._read_raw(response_len), 2)    # type: bytearray
            except asyncio.TimeoutError:    # pragma: no cover
                self.warning_log("Failed to read %s bytes from Spike", response_len)
                return None

            if response[-1] != 0:
                self.info_log("Bridge Status: %s != 0", response[-1])
            if self.config['verify_checksums_on_readback'] and self._checksum(response[0:-1]) != 0:   # pragma: no cover
                self.warning_log("Checksum mismatch for response: %s", "".join("%02x " % b for b in response))
                # we resync by flushing the input
                self._writer.transport.serial.reset_input_buffer()
                # pylint: disable-msg=protected-access
                self._reader._buffer = bytearray()
                return None

            return response

    def _create_cmd_str(self, node, cmd, data):
        if node > 15:
            raise AssertionError("Node must be 0-15.")
        cmd_str = bytearray()
        cmd_str.append((8 << 4) + node)
        cmd_str.append(len(data) + 2)
        cmd_str.append(cmd)
        cmd_str.extend(data)
        cmd_str.append(self._checksum(cmd_str))
        cmd_str.append(0)
        return cmd_str

    def _get_wait_ms_for_cmd(self, cmd):
        if (cmd & 0xF0) == 0x80:
            # special case for LED updates
            return self.config['wait_times'][0x80] if 0x80 in self.config['wait_times'] else 0

        return self.config['wait_times'][cmd] if cmd in self.config['wait_times'] else 0

    async def send_cmd_sync(self, node, cmd, data):
        """Send cmd which does not require a response."""
        cmd_str = self._create_cmd_str(node, cmd, data)
        wait_ms = self._get_wait_ms_for_cmd(cmd)
        await self._send_cmd_without_response(cmd_str, wait_ms)

    async def _send_cmd_without_response(self, cmd_str, wait_ms):
        async with self._bus_write:
            await self._send_raw(cmd_str)
            if wait_ms:
                await self._send_raw(bytearray([1, wait_ms]))

    async def send_cmd_raw(self, data, wait_ms=0):
        """Send raw command."""
        async with self._bus_write:
            await self._send_raw(data)
            if wait_ms:
                await self._send_raw(bytearray([1, wait_ms]))

    def send_cmd_async(self, node, cmd, data):
        """Send cmd which does not require a response."""
        cmd_str = self._create_cmd_str(node, cmd, data)
        wait_ms = self._get_wait_ms_for_cmd(cmd)
        # queue command
        self._cmd_queue.put_nowait((cmd_str, wait_ms))

    def send_cmd_raw_async(self, data, wait_ms=0):
        """Send raw cmd which does not require a response."""
        # queue command
        self._cmd_queue.put_nowait((data, wait_ms))

    def _read_inputs(self, node):
        if self.node_firmware_version[node] >= 0x2800:
            # in node firmware 0.40.0 and newer the response uses 12 bytes (10 payload)
            return self.send_cmd_and_wait_for_response(node, SpikeNodebus.GetInputState, bytearray(), 12)

        # older firmware used 10 bytes responses (8 payload)
        return self.send_cmd_and_wait_for_response(node, SpikeNodebus.GetInputState, bytearray(), 10)

    @staticmethod
    def _input_to_int(state):
        if state is False or state is None or len(state) < 8:
            return 0

        result = 0
        for i in range(8):
            result += pow(256, i) * int(state[i])

        return result

    def configure_dmd(self):
        """Configure a DMD."""
        if self.dmd:
            raise AssertionError("Can only configure dmd once.")
        self.dmd = SpikeDMD(self)
        return self.dmd

    async def configure_stepper(self, number: str, config: dict) -> "StepperPlatformInterface":
        """Configure a stepper in Spike."""
        # TODO: validate that light number is not used in a stepper and a light
        return SpikeStepper(number, config, self)

    @classmethod
    def get_stepper_config_section(cls):
        """Return config validator name."""
        return "spike_stepper_settings"

    async def _init_bridge(self):
        # send ctrl+c to stop whatever is running
        self.debug_log("Resetting console")
        self._writer.write(b'\x03reset\n')
        # wait for the serial
        await asyncio.sleep(.1)
        # flush input
        self._writer.transport.serial.reset_input_buffer()
        # pylint: disable-msg=protected-access
        self._reader._buffer = bytearray()
        # start mpf-spike-bridge
        self.debug_log("Starting MPF bridge")
        if self.config['bridge_debug']:
            log_file = self.config['bridge_debug_log']
            binary = "RUST_BACKTRACE=full {}".format(self.config['bridge_path'])
        else:
            log_file = "/dev/null"
            binary = self.config['bridge_path']

        if self.config['spike_version'] == "1":
            spike_version = "SPIKE1"
        else:
            spike_version = "SPIKE2"
        self._writer.write("{} {} {} 2>{}\r\n".format(binary, self.config['runtime_baud'], spike_version,
                                                      log_file).encode())

        welcome_str = b'MPF Spike Bridge!'
        await asyncio.sleep(.1)
        # read until first capital M
        while True:
            byte = await self._reader.readexactly(1)
            if ord(byte) == welcome_str[0]:
                break

        data = await self._reader.read(100)
        if data[:len(welcome_str) - 1] != welcome_str[1:]:
            raise AssertionError("Expected '{}' got '{}'".format(welcome_str[1:], data[:len(welcome_str) - 1]))
        self.debug_log("Bridge started")

        if self.config['runtime_baud']:
            # increase baud rate
            self.debug_log("Increasing baudrate to %s", self.config['runtime_baud'])
            self._writer.transport.serial.baudrate = self.config['runtime_baud']

        await asyncio.sleep(.1)
        self._reader._buffer = bytearray()

    # pylint: disable-msg=too-many-statements
    # pylint: disable-msg=too-many-branches
    async def _initialize(self) -> None:    # noqa
        await self._init_bridge()

        self.debug_log("Resetting node bus and configuring traffic.")
        await self.send_cmd_sync(0, SpikeNodebus.Reset, bytearray())
        # wait 3s (same as spike)
        for _ in range(12):
            await self._send_raw(bytearray([1, 250]))

        await self.send_cmd_sync(0, SpikeNodebus.SetTraffic, bytearray([34]))  # block traffic (false)
        await self.send_cmd_sync(0, SpikeNodebus.SetTraffic, bytearray([17]))  # set traffic

        initialized_nodes = {0}

        while True:
            # poll to iterate nodes
            await self.send_cmd_raw([0])
            node_str = await self._read_raw(1)
            if node_str is None:
                self.warning_log("Initial poll timeouted")
                await asyncio.sleep(.5)
                continue

            node = node_str[0]
            if node == 0:
                # all nodes initialised
                break

            if node == 0xF0:
                # local switches. just read them to make them go away
                await self._read_inputs(0)
                continue

            initialized_nodes.add(node)
            self.debug_log("Poll nodes: %s", node)

            await self.send_cmd_sync(node, SpikeNodebus.SetTraffic, bytearray([16]))  # clear traffic
            await self.send_cmd_sync(node, SpikeNodebus.SetTraffic, bytearray([32]))  # block traffic (true)

            if node not in self._nodes:
                self.warning_log("Found a node %s during initial polling which is not configured", node)

        if set(self._nodes) - initialized_nodes:
            self.warning_log("Not all nodes found during init. Missing %s Found: %s",
                             set(self._nodes) - initialized_nodes, initialized_nodes)

        await self.send_cmd_sync(0, SpikeNodebus.SetTraffic, bytearray([34]))  # block traffic (false)

        # get bridge version
        await self.send_cmd_raw([SpikeNodebus.GetBridgeVersion, 0, 3], 0)
        bridge_version = await self._read_raw(3)
        self.debug_log("Bridge version: %s", "".join("0x%02x " % b for b in bridge_version))

        # get bridge status
        await self.send_cmd_raw([SpikeNodebus.GetBridgeStatus, 0, 1], 0)
        bridge_status = await self._read_raw(1)
        self.debug_log("Bridge status: %s", "".join("0x%02x " % b for b in bridge_status))

        for node in self._nodes:
            if node == 0:
                continue
            self.debug_log("GetVersion on node %s", node)
            fw_version = await self.send_cmd_and_wait_for_response(node, SpikeNodebus.GetVersion, bytearray(), 12)
            if not fw_version:
                self.warning_log("Did not get version for node: %s. Ignoring node.", node)
                continue
            self.info_log("Node: %s Firmware: %s.%s.%s", node, fw_version[1], fw_version[2], fw_version[3])
            if fw_version[0] != node:
                self.warning_log("Node: %s Version Response looks bogus (node ID does not match): %s",
                                 node, "".join("0x%02x " % b for b in fw_version))

            # we need this to calculate the right times for this node
            self.ticks_per_sec[node] = (fw_version[9] << 8) + fw_version[8]

            # remember firmware version of node
            self.node_firmware_version[node] = fw_version[1] << 16 | fw_version[2] << 8 | fw_version[3]

            self.debug_log("GetFullID on node %s", node)
            full_board_id_0 = await self.send_cmd_and_wait_for_response(node, SpikeNodebus.GetFullBoardId,
                                                                        bytearray([0]), 18)
            if not full_board_id_0:
                self.warning_log("Did not get full_board_id 0 for node: %s.", node)

            full_board_id_1 = await self.send_cmd_and_wait_for_response(node, SpikeNodebus.GetFullBoardId,
                                                                        bytearray([1]), 18)
            if not full_board_id_1:
                self.warning_log("Did not get full_board_id 1 for node: %s.", node)

            if full_board_id_0 and full_board_id_1:
                self.info_log("Node %s: Full Board ID: %s %s",
                              node, "".join("0x%02x " % b for b in full_board_id_0),
                              "".join("0x%02x " % b for b in full_board_id_1))

            # Set response time (redundant but send multiple times)
            # wait time based on the baud rate of the bus: (460800 * 0x98852841 * 200) >> 0x30 = 0x345
            response_time = self.config['response_time']
            await self.send_cmd_raw([SpikeNodebus.SetResponseTime, 0x02, int(response_time & 0xff),
                                     int(response_time >> 8), 0], 0)

            self.debug_log("GetChecksum on node %s", node)
            checksum = await self.send_cmd_and_wait_for_response(node, SpikeNodebus.GetChecksum,
                                                                 bytearray([0xff, 0x00]), 4)
            if checksum and checksum[0] == 0 and checksum[1] == 0 and checksum[2] == 0 and checksum[3] == 0:
                self.info_log("Checksum good for node %s", node)
            elif checksum:
                self.warning_log("Checksum %s for node %s is != 0000. This might indicate a broken firmware.",
                                 "".join("0x%02x " % b for b in checksum), node)
            else:
                self.warning_log("Did not get checksum for node %s", node)

        for node in self._nodes:
            self.debug_log("Initial read inputs on node %s", node)
            initial_inputs = await self._read_inputs(node)
            self._inputs[node] = self._input_to_int(initial_inputs)

        for node in self._nodes:
            if node == 0:
                continue
            self.debug_log("GetStatus and GetCoilCurrent on node %s", node)
            node_status = await self.send_cmd_and_wait_for_response(node, SpikeNodebus.GetStatus, bytearray(), 10)
            if node_status:
                self.debug_log("Node: %s Status: %s", node, "".join("0x%02x " % b for b in node_status))
            else:
                self.warning_log("Did not get status for node %s", node)

            if self.node_firmware_version[node] >= 0x3100:
                self.debug_log("SetLEDMask, CoilSetMask, CoilSetOCTime, CoilSetOCBehavior, and SetNumLEDsInputs "
                               "on node %s", node)

                node_status = await self.send_cmd_and_wait_for_response(node, SpikeNodebus.GetStatus, bytearray(), 10)
                if node_status:
                    self.debug_log("Node: %s Status: %s", node, "".join("0x%02x " % b for b in node_status))
                else:
                    self.warning_log("Did not get status for node %s", node)

                # set 96 leds and 60 inputs during OC detection
                await self.send_cmd_sync(node, SpikeNodebus.SetNumLEDsInputs, bytearray([0x60, 0, 0x40, 0]))

                # mask out all coils
                await self.send_cmd_sync(node, SpikeNodebus.CoilSetMask, bytearray([0xff, 0x01]))

                # mask out all leds
                await self.send_cmd_sync(node, SpikeNodebus.SetLEDMask, bytearray([0xff] * 12))

                # 5ms oc detect
                oc_time = int(5 * self.ticks_per_sec[node] / 1000)
                await self.send_cmd_sync(node, SpikeNodebus.CoilSetOCTime,
                                         bytearray([oc_time & 0xff, (oc_time >> 8) & 0xff]))

                for coil in range(0, 8):
                    await self.send_cmd_sync(node, SpikeNodebus.CoilSetMask, bytearray([0xff - (1 << coil), 0x01]))
                    # pulse coils
                    await self.send_cmd_sync(node, SpikeNodebus.CoilFireRelease,
                                             bytearray([coil, 0xff, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]))
                    node_status = await self.send_cmd_and_wait_for_response(node, SpikeNodebus.GetStatus, bytearray(),
                                                                            10)
                    if node_status:
                        self.debug_log("Node: %s Status: %s", node, "".join("0x%02x " % b for b in node_status))
                    else:
                        self.warning_log("Did not get status for node %s", node)

                # enable all coils
                await self.send_cmd_sync(node, SpikeNodebus.CoilSetMask, bytearray([0, 0]))

                # enable all leds
                await self.send_cmd_sync(node, SpikeNodebus.SetLEDMask, bytearray([0] * 12))

                # 100ms oc time
                oc_time = int(self.config['oc_time'] * self.ticks_per_sec[node] / 1000)
                await self.send_cmd_sync(node, SpikeNodebus.CoilSetOCTime,
                                         bytearray([oc_time & 0xff, (oc_time >> 8) & 0xff]))
                # set whatever spike sets
                await self.send_cmd_sync(node, SpikeNodebus.CoilSetOCBehavior, bytearray([0x01]))

                # set number of LEDs and inputs based on node config
                if node in self.config['node_config'] and self.config['node_config'][node]["num_leds"] is not None and \
                        self.config['node_config'][node]["num_inputs"] is not None:
                    await self.send_cmd_sync(node, SpikeNodebus.SetNumLEDsInputs,
                                             bytearray([self.config['node_config'][node]["num_leds"], 0,
                                                        self.config['node_config'][node]["num_inputs"], 0]))

            await self.send_cmd_and_wait_for_response(node, SpikeNodebus.GetCoilCurrent, bytearray([0]), 12)

        for node, config in self.config['node_config'].items():
            if self.node_firmware_version[node] >= 0x3100 and config['coil_priorities']:
                # configure coil priorities
                priority_response = await self.send_cmd_and_wait_for_response(
                    node, SpikeNodebus.CoilSetPriority,
                    bytearray([len(config['coil_priorities'])] + config['coil_priorities'] +
                              [len(config['coil_priorities'])]), 3)

                if not priority_response:
                    self.warning_log("Failed to set coil priority on node %s", node)

        self.debug_log("Configuring traffic.")
        await self.send_cmd_sync(0, SpikeNodebus.SetTraffic, bytearray([0x11]))  # set traffic

        self.info_log("SPIKE init done.")



from mpf.tests.MpfTestCase import MpfTestCase, test_config
from unittest.mock import MagicMock


class TestToughEntranceSwitch(MpfTestCase):
    def get_config_file(self):
        return 'trough_entrance_switch_initial_balls.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def test_initial_balls(self):
        """On startup the machine should count three balls."""
        self.assertNumBallsKnown(3)

    @test_config("trough_entrance_switch.yaml")
    def test_ball_settling(self):
        """Test that the trough waits for a ball to settle before ejecting."""
        self.assertNumBallsKnown(0)

        # request a ball to the playfield
        self.machine.playfield.add_ball()

        # add a ball to drain
        self.hit_switch_and_run("s_drain", .1)
        self.assertEqual("disabled", self.machine.coils["c_drain_eject"].hw_driver.state)

        # it waits for the ball to settle and ejects
        self.advance_time_and_run(.5)
        self.assertEqual("pulsed_20", self.machine.coils["c_drain_eject"].hw_driver.state)

        # ball leaves drain switch
        self.release_switch_and_run("s_drain", .5)

        # and enters the trough
        self.assertEqual("disabled", self.machine.coils["c_trough_release"].hw_driver.state)
        self.hit_and_release_switch("s_trough_enter")
        self.advance_time_and_run(.5)

        # it waits for the ball to settle
        self.assertEqual("disabled", self.machine.coils["c_trough_release"].hw_driver.state)
        self.advance_time_and_run(3)
        self.assertEqual("pulsed_20", self.machine.coils["c_trough_release"].hw_driver.state)



from mpf.tests.MpfTestCase import MpfTestCase, MagicMock


class TestStepStick(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/step_stick/'

    def get_platform(self):
        # no force platform. we are testing step stick
        return False

    def test_step_stick(self):
        direction = self.machine.digital_outputs["c_direction"].hw_driver
        step_enable = self.machine.digital_outputs["c_step"].hw_driver.enable = MagicMock()
        step_disable = self.machine.digital_outputs["c_step"].hw_driver.disable = MagicMock()
        enable = self.machine.digital_outputs["c_enable"].hw_driver
        self.assertEqual("enabled", direction.state)
        self.assertEqual("enabled", enable.state)

        self.advance_time_and_run(1)

        self.assertEqual(25, step_enable.call_count)
        self.assertEqual(25, step_disable.call_count)

        self.hit_switch_and_run("s_home", .1)
        step_enable = self.machine.digital_outputs["c_step"].hw_driver.enable = MagicMock()
        step_disable = self.machine.digital_outputs["c_step"].hw_driver.disable = MagicMock()

        self.advance_time_and_run(1)
        self.assertEqual(0, step_enable.call_count)
        self.assertEqual(0, step_disable.call_count)

        self.post_event("test_01")
        self.advance_time_and_run(1)
        self.assertEqual("enabled", direction.state)
        self.assertEqual(20, step_enable.call_count)
        self.assertEqual(20, step_disable.call_count)

        step_enable = self.machine.digital_outputs["c_step"].hw_driver.enable = MagicMock()
        step_disable = self.machine.digital_outputs["c_step"].hw_driver.disable = MagicMock()
        self.post_event("test_00")
        self.advance_time_and_run(1)
        self.assertEqual("disabled", direction.state)
        self.assertEqual(10, step_enable.call_count)
        self.assertEqual(10, step_disable.call_count)

        step_enable = self.machine.digital_outputs["c_step"].hw_driver.enable = MagicMock()
        step_disable = self.machine.digital_outputs["c_step"].hw_driver.disable = MagicMock()
        self.post_event("test_10")
        self.advance_time_and_run(2)
        self.assertEqual("enabled", direction.state)
        self.assertEqual(40, step_enable.call_count)
        self.assertEqual(40, step_disable.call_count)



"""Contains code for an FadeCandy hardware for RGB LEDs."""
from typing import List

import logging
import json
import struct

from mpf.core.utility_functions import Util
from mpf.platforms.openpixel import OpenPixelClient
from mpf.platforms.openpixel import OpenpixelHardwarePlatform

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import


class FadecandyHardwarePlatform(OpenpixelHardwarePlatform):

    """Base class for the FadeCandy hardware platform."""

    __slots__ = []  # type: List[str]

    def __init__(self, machine: "MachineController") -> None:
        """Initialise Fadecandy.

        Args:
        ----
            machine: The main ``MachineController`` object.
        """
        super().__init__(machine)

        self.log = logging.getLogger("FadeCandy")
        self.log.debug("Configuring FadeCandy hardware interface.")

    def __repr__(self):
        """Return string representation."""
        return '<Platform.FadeCandy>'

    async def _setup_opc_client(self):
        self.opc_client = FadeCandyOPClient(self.machine, self.machine.config['open_pixel_control'])
        await self.opc_client.connect()


class FadeCandyOPClient(OpenPixelClient):

    """Base class of an OPC client which connects to a FadeCandy server.

    This class implements some FadeCandy-specific features that are not
    available with generic OPC implementations.

    """

    __slots__ = ["gamma", "whitepoint", "linear_slope", "linear_cutoff", "keyframe_interpolation", "dithering",
                 "config"]

    def __init__(self, machine, config):
        """Initialise Fadecandy client.

        Args:
        ----
            machine: The main ``MachineController`` instance.
            config: Dictionary which contains configuration settings for the
                OPC client.
        """
        super().__init__(machine, config)

        self.log = logging.getLogger('FadeCandyClient')

        self.update_every_tick = True

        self.config = self.machine.config_validator.validate_config('fadecandy',
                                                                    self.machine.config['fadecandy'])

        self.gamma = self.config['gamma']
        self.whitepoint = Util.string_to_event_list(self.config['whitepoint'])

        self.whitepoint[0] = float(self.whitepoint[0])
        self.whitepoint[1] = float(self.whitepoint[1])
        self.whitepoint[2] = float(self.whitepoint[2])

        self.linear_slope = self.config['linear_slope']
        self.linear_cutoff = self.config['linear_cutoff']
        self.keyframe_interpolation = self.config['keyframe_interpolation']
        self.dithering = self.config['dithering']

        if not self.keyframe_interpolation:
            self.update_every_tick = False

    async def connect(self):
        """Connect to the hardware."""
        await super().connect()
        self.set_global_color_correction()
        self.write_firmware_options()

    def __repr__(self):
        """Return str representation."""
        return '<Platform.FadeCandyOPClient>'

    def set_global_color_correction(self):
        """Write the current global color correction settings to the FadeCandy server.

        This includes gamma, white point, linear slope, and linear cutoff.
        """
        msg = json.dumps({
            'gamma': self.gamma,
            'whitepoint': self.whitepoint,
            'linearSlope': self.linear_slope,
            'linearCutoff': self.linear_cutoff
        })

        self.send(struct.pack(
            "!BBHHH", 0x00, 0xFF, len(msg) + 4, 0x0001, 0x0001) + bytes(msg, 'UTF-8'))

    def write_firmware_options(self):
        """Write the current firmware settings (keyframe interpolation and dithering) to the FadeCandy hardware."""
        config_byte = 0x00

        if not self.dithering:
            config_byte |= 0x01

        if not self.keyframe_interpolation:
            config_byte |= 0x02

        # manual LED control
        # config_byte = config_byte | 0x04

        # turn LED on
        # config_byte = config_byte | 0x08

        self.send(struct.pack(
            "!BBHHHB", 0x00, 0xFF, 0x0005, 0x0001, 0x0002, config_byte))



"""Runtime error in MPF or MPF-MC."""
from mpf.exceptions.base_error import BaseError


class MpfRuntimeError(BaseError):

    """Runtime error."""

    def get_short_name(self):
        """Return short name."""
        return "RE"

    def get_long_name(self):
        """Return long name."""
        return "Runtime Error"



"""Platform to use lights connected to a driver."""
from mpf.platforms.interfaces.light_platform_interface import LightPlatformSoftwareFade, LightPlatformInterface
from mpf.core.platform import LightsPlatform


class DriverLight(LightPlatformSoftwareFade):

    """A coil which is used to drive a light."""

    __slots__ = ["driver"]

    def __init__(self, number, driver, loop, software_fade_ms):
        """Initialise coil as light."""
        super().__init__(number, loop, software_fade_ms)
        self.driver = driver

    def set_brightness(self, brightness: float):
        """Set pwm to coil."""
        if brightness <= 0:
            self.driver.disable()
        else:
            self.driver.enable(hold_power=brightness)

    def get_board_name(self):
        """Return board name of underlaying driver."""
        return self.driver.hw_driver.get_board_name()

    def is_successor_of(self, other):
        """Not possible."""
        raise AssertionError("Not possible in DriverLights.")

    def get_successor_number(self):
        """Not possible."""
        raise AssertionError("Not possible in DriverLights.")

    def __lt__(self, other):
        """Order lights by string."""
        return self.number < other.number


class DriverLightPlatform(LightsPlatform):

    """Lights on drivers."""

    __slots__ = ["_lights"]

    def __init__(self, machine):
        """Initialize platform."""
        super().__init__(machine)
        self._lights = []
        self.features["tickless"] = True

    def stop(self):
        """Stop all fades."""
        for light in self._lights:
            light.stop()
        self._lights = []

    def configure_light(self, number: str, subtype: str, config, platform_settings: dict) -> LightPlatformInterface:
        """Configure a light on a driver."""
        del config
        driver = DriverLight(number.strip(), self.machine.coils[number.strip()], self.machine.clock.loop,
                             int(1 / self.machine.config['mpf']['default_light_hw_update_hz'] * 1000))
        self._lights.append(driver)
        return driver

    def parse_light_number_to_channels(self, number: str, subtype: str):
        """Parse number."""
        del subtype
        channel_list = [
            {
                "number": number,
                "platform": "drivers"
            }
        ]
        return channel_list



"""Transfer a ball between two playfields.

E.g. lower to upper playfield via a ramp.
"""
from mpf.core.events import event_handler
from mpf.core.system_wide_device import SystemWideDevice


class PlayfieldTransfer(SystemWideDevice):

    """Device which move a ball from one playfield to another."""

    config_section = 'playfield_transfers'
    collection = 'playfield_transfers'
    class_label = 'playfield_transfer'

    __slots__ = ["target", "source"]

    def __init__(self, machine, name):
        """Initialise playfield transfer."""
        self.target = None
        self.source = None
        super().__init__(machine, name)

    async def _initialize(self):
        await super()._initialize()
        if self.config['ball_switch']:
            self.machine.events.add_handler('init_phase_3',
                                            self._configure_switch)

        # load target playfield
        self.target = self.config['eject_target']
        self.source = self.config['captures_from']

    def _configure_switch(self, **kwargs):
        del kwargs
        self.machine.switch_controller.add_switch_handler_obj(
            switch=self.config['ball_switch'],
            callback=self.transfer,
            state=1, ms=0)

    @event_handler(1)
    def event_transfer(self, **kwargs):
        """Event handler for transfer event."""
        del kwargs
        self.transfer()

    def transfer(self):
        """Transfer a ball to the target playfield."""
        self.debug_log("Ball went from %s to %s", self.source.name,
                       self.target.name)

        # source playfield is obviously active
        # we will continue using a callback to keep the ball count sane
        # (otherwise it may go to -1 during the next event)
        self.machine.events.post('sw_' + self.source.name + '_active',
                                 callback=self._ball_went_through2, balls=1)
        # event docstring in playfield module.

        self.machine.events.post('playfield_transfer_{}_ball_transferred'.format(self.name),
                                 source=self.source, target=self.target)
        '''event: playfield_transfer_(playfield_transfer)_ball_transferred
        desc: The playfield_transfer called (playfield_transfer) transferred a ball from playfield (source) to
        playfield (target).

        args:
        source: The source playfield.
        target: The target playfield.
        '''

    # used as callback in _ball_went_through
    def _ball_went_through2(self, **kwargs):
        del kwargs
        # trigger remove ball from source playfield
        self.machine.events.post(
            'balldevice_captured_from_' + self.source.name,
            balls=1)
        # event docstring covered elsewhere

        # inform target playfield about incoming ball
        self.machine.events.post(
            'balldevice_' + self.name + '_ejecting_ball',
            balls=1,
            target=self.target,
            timeout=0,
            callback=self._ball_went_through3)
        # event docstring covered elsewhere

    # used as callback in _ball_went_through2
    def _ball_went_through3(self, balls, target, timeout):
        del balls
        del target
        del timeout
        # promise (and hope) that it actually goes there
        self.machine.events.post(
            'balldevice_' + self.name + '_ball_eject_success',
            balls=1,
            target=self.target,
            callback=self._ball_went_through4)
        self.target.available_balls += 1
        # event docstring covered elsewhere

    # used as callback in _ball_went_through3
    def _ball_went_through4(self, balls, target):
        del balls
        del target
        # since we confirmed eject target playfield has to be active
        self.machine.events.post('sw_' + self.target.name + '_active')
        # event docstring covered elsewhere



"""Shot profiles."""
from typing import List

from mpf.core.mode import Mode

from mpf.core.system_wide_device import SystemWideDevice

from mpf.core.mode_device import ModeDevice


class ShotProfile(ModeDevice, SystemWideDevice):

    """A shot profile."""

    config_section = 'shot_profiles'
    collection = 'shot_profiles'
    class_label = 'shot_profile'

    __slots__ = []  # type: List[str]

    def device_removed_from_mode(self, mode: Mode) -> None:
        """Remove from mode."""



"""Defines for OPP platform."""
from typing import List


class OppRs232Intf:

    """Constants for OPP serial protocol."""

    __slots__ = []  # type: List[str]

    GET_SER_NUM_CMD = b'\x00'
    GET_PROD_ID_CMD = b'\x01'
    GET_VERS_CMD = b'\x02'
    SET_SER_NUM_CMD = b'\x03'
    RESET_CMD = b'\x04'
    GO_BOOT_CMD = b'\x05'
    CFG_SOL_CMD = b'\x06'
    KICK_SOL_CMD = b'\x07'
    READ_GEN2_INP_CMD = b'\x08'
    CFG_INP_CMD = b'\x09'
    SAVE_CFG_CMD = b'\x0b'
    ERASE_CFG_CMD = b'\x0c'
    GET_GEN2_CFG = b'\x0d'
    SET_GEN2_CFG = b'\x0e'
    CHNG_NEO_CMD = b'\x0f'
    CHNG_NEO_COLOR = b'\x10'
    CHNG_NEO_COLOR_TBL = b'\x11'
    SET_NEO_COLOR_TBL = b'\x12'
    INCAND_CMD = b'\x13'
    CFG_IND_SOL_CMD = b'\x14'
    CFG_IND_INP_CMD = b'\x15'
    SET_IND_NEO_CMD = b'\x16'
    SET_SOL_INP_CMD = b'\x17'
    UPGRADE_OTHER_BRD = b'\x18'
    READ_MATRIX_INP = b'\x19'

    INV_CMD = b'\xf0'
    ILLEGAL_CMD = b'\xfe'
    EOM_CMD = b'\xff'

    CARD_ID_TYPE_MASK = b'\xf0'
    CARD_ID_GEN2_CARD = b'\x20'

    NUM_G2_WING_PER_BRD = 4
    WING_SOL = b'\x01'
    WING_INP = b'\x02'
    WING_INCAND = b'\x03'
    WING_SW_MATRIX_OUT = b'\x04'
    WING_SW_MATRIX_IN = b'\x05'
    WING_NEO = b'\x06'
    WING_HI_SIDE_INCAND = b'\x07'
    WING_NEO_SOL = b'\x08'
    WING_SW_MATRIX_OUT_LOW_WING = b'\x0a'
    WING_LAMP_MATRIX_COL_WING = b'\x0b'
    WING_LAMP_MATRIX_ROW_WING = b'\x0c'

    NUM_G2_INP_PER_BRD = 32
    CFG_INP_STATE = b'\x00'
    CFG_INP_FALL_EDGE = b'\x01'
    CFG_INP_RISE_EDGE = b'\x02'

    # Solenoid configuration constants
    CFG_BYTES_PER_SOL = 3
    INIT_KICK_OFFSET = 1
    DUTY_CYCLE_OFFSET = 2
    NUM_G2_SOL_PER_BRD = 16

    CFG_SOL_DISABLE = b'\x00'
    CFG_SOL_USE_SWITCH = b'\x01'
    CFG_SOL_AUTO_CLR = b'\x02'
    CFG_SOL_ON_OFF = b'\x04'
    CFG_SOL_DLY_KICK = b'\x08'
    CFG_SOL_USE_MTRX_INP = b'\x10'
    CFG_SOL_CAN_CANCEL = b'\x20'

    CFG_SOL_INP_REMOVE = b'\x80'

    NUM_COLOR_TBL = 32
    NEO_CMD_ON = 0x80
    SERIAL_LED_CMD_FADE = 0x40

    INCAND_ROT_LEFT = b'\x00'
    INCAND_ROT_RIGHT = b'\x01'
    INCAND_LED_ON = b'\x02'
    INCAND_LED_OFF = b'\x03'
    INCAND_BLINK_SLOW = b'\x04'
    INCAND_BLINK_FAST = b'\x05'
    INCAND_BLINK_OFF = b'\x06'
    INCAND_SET_ON_OFF = b'\x07'

    INCAND_SET_CMD = b'\x80'
    INCAND_SET_ON = b'\x01'
    INCAND_SET_BLINK_SLOW = b'\x02'
    INCAND_SET_BLINK_FAST = b'\x04'

    CRC8_LOOKUP = [
        0x00, 0x07, 0x0e, 0x09, 0x1c, 0x1b, 0x12, 0x15, 0x38, 0x3f, 0x36, 0x31, 0x24, 0x23, 0x2a, 0x2d,
        0x70, 0x77, 0x7e, 0x79, 0x6c, 0x6b, 0x62, 0x65, 0x48, 0x4f, 0x46, 0x41, 0x54, 0x53, 0x5a, 0x5d,
        0xe0, 0xe7, 0xee, 0xe9, 0xfc, 0xfb, 0xf2, 0xf5, 0xd8, 0xdf, 0xd6, 0xd1, 0xc4, 0xc3, 0xca, 0xcd,
        0x90, 0x97, 0x9e, 0x99, 0x8c, 0x8b, 0x82, 0x85, 0xa8, 0xaf, 0xa6, 0xa1, 0xb4, 0xb3, 0xba, 0xbd,
        0xc7, 0xc0, 0xc9, 0xce, 0xdb, 0xdc, 0xd5, 0xd2, 0xff, 0xf8, 0xf1, 0xf6, 0xe3, 0xe4, 0xed, 0xea,
        0xb7, 0xb0, 0xb9, 0xbe, 0xab, 0xac, 0xa5, 0xa2, 0x8f, 0x88, 0x81, 0x86, 0x93, 0x94, 0x9d, 0x9a,
        0x27, 0x20, 0x29, 0x2e, 0x3b, 0x3c, 0x35, 0x32, 0x1f, 0x18, 0x11, 0x16, 0x03, 0x04, 0x0d, 0x0a,
        0x57, 0x50, 0x59, 0x5e, 0x4b, 0x4c, 0x45, 0x42, 0x6f, 0x68, 0x61, 0x66, 0x73, 0x74, 0x7d, 0x7a,
        0x89, 0x8e, 0x87, 0x80, 0x95, 0x92, 0x9b, 0x9c, 0xb1, 0xb6, 0xbf, 0xb8, 0xad, 0xaa, 0xa3, 0xa4,
        0xf9, 0xfe, 0xf7, 0xf0, 0xe5, 0xe2, 0xeb, 0xec, 0xc1, 0xc6, 0xcf, 0xc8, 0xdd, 0xda, 0xd3, 0xd4,
        0x69, 0x6e, 0x67, 0x60, 0x75, 0x72, 0x7b, 0x7c, 0x51, 0x56, 0x5f, 0x58, 0x4d, 0x4a, 0x43, 0x44,
        0x19, 0x1e, 0x17, 0x10, 0x05, 0x02, 0x0b, 0x0c, 0x21, 0x26, 0x2f, 0x28, 0x3d, 0x3a, 0x33, 0x34,
        0x4e, 0x49, 0x40, 0x47, 0x52, 0x55, 0x5c, 0x5b, 0x76, 0x71, 0x78, 0x7f, 0x6a, 0x6d, 0x64, 0x63,
        0x3e, 0x39, 0x30, 0x37, 0x22, 0x25, 0x2c, 0x2b, 0x06, 0x01, 0x08, 0x0f, 0x1a, 0x1d, 0x14, 0x13,
        0xae, 0xa9, 0xa0, 0xa7, 0xb2, 0xb5, 0xbc, 0xbb, 0x96, 0x91, 0x98, 0x9f, 0x8a, 0x8d, 0x84, 0x83,
        0xde, 0xd9, 0xd0, 0xd7, 0xc2, 0xc5, 0xcc, 0xcb, 0xe6, 0xe1, 0xe8, 0xef, 0xfa, 0xfd, 0xf4, 0xf3]

    @staticmethod
    def calc_crc8_whole_msg(msg_chars):
        """Calculate CRC for message."""
        crc8_byte = 0xff
        for ind_char in msg_chars:
            ind_int = ind_char
            crc8_byte = OppRs232Intf.CRC8_LOOKUP[crc8_byte ^ ind_int]
        return bytes([crc8_byte])

    @staticmethod
    def calc_crc8_part_msg(msg_chars, start_index, num_chars):
        """Calculate CRC for part of a message."""
        crc8_byte = 0xff
        index = 0
        if len(msg_chars) < start_index + num_chars:
            raise AssertionError("String too short for {} chars of CRC: {}". format(
                num_chars,
                "".join(" 0x%02x" % b for b in msg_chars[start_index:])))
        while index < num_chars:
            ind_int = msg_chars[start_index + index]
            crc8_byte = OppRs232Intf.CRC8_LOOKUP[crc8_byte ^ ind_int]
            index += 1
        return bytes([crc8_byte])



"""Test digital score reels."""

from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase

class TestDigitalScoreReels(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'test_digital_score_reels.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/digital_score_reels'

    def test_player_score(self):
        self.mock_event('score_reel_player_score_player1')
        self.mock_event('score_reel_player_score_player2')
        self.start_two_player_game()
        self.assertGameIsRunning()

        self.assertPlayerNumber(1)
        self.advance_time_and_run()
        self.machine.game.player.score = 123
        self.advance_time_and_run()
        self.assertEqual({"1": "20", "2": "2", "3": "4", "4": "6"}, self._last_event_kwargs['score_reel_player_score_player1'])
        self.assertEventNotCalled('score_reel_player_score_player2')

        self.drain_all_balls()
        self.assertPlayerNumber(2)
        self.machine.game.player.score = 9876
        self.advance_time_and_run()
        self.assertEqual({"1": "18", "2": "16", "3": "14", "4": "12"}, self._last_event_kwargs['score_reel_player_score_player2'])

    def test_arbitrary_event(self):
        self.mock_event('score_reel_arbitrary_event')
        self.post_event_with_params('arbitrary_event', value='AB')
        self.advance_time_and_run()
        self.assertEqual({"1": "6", "2": "1", "3": "2"}, self._last_event_kwargs['score_reel_arbitrary_event'])

        self.post_event_with_params('arbitrary_event', value='DXB')
        self.advance_time_and_run()
        self.assertEqual({"1": "4", "2":"6", "3": "2"}, self._last_event_kwargs['score_reel_arbitrary_event'])

        self.post_event_with_params('arbitrary_event', value='EC')
        self.advance_time_and_run()
        self.assertEqual({"1": "6", "2":"5", "3": "3"}, self._last_event_kwargs['score_reel_arbitrary_event'])


"""Contains the Player class which represents a player in a pinball game."""
import copy
import logging

from mpf.core.utility_functions import Util


class Player:

    """Base class for a player in a game.

    One instance of this class is automatically created for each player.

    The game mode maintains a ``player`` attribute which always points to the
    current player and is available via ``self.machine.game.player``.

    It also contains a ``player_list`` attribute which is a list
    of the player instances (in order) which you can use to access the
    non-current player.

    This Player class is responsible for tracking *player variables* which
    is a dictionary of key/value pairs maintained on a per-player basis. There
    are several ways they can be used:

    First, player variables can be accessed as attributes of the player
    object directly. For example, to set a player variable `foo` for the
    current player, you could use:

    .. code::

        self.machine.player.foo = 0

    If that variable didn't exist, it will be automatically created.

    You can get the value of player variables by accessing them directly. For
    example:

    .. code::

        print(self.machine.player.foo)  # prints 0

    If you attempt to access a player variable that doesn't exist, it will
    automatically be created with a value of ``0``.

    Every time a player variable is created or changed, an MPF event is posted
    in the form *player_* plus the variable name. For example, creating or
    changing the `foo` variable will cause an event called *player_foo* to
    be posted.

    The player variable event will have four parameters posted along with it:

    * ``value`` (the new value)
    * ``prev_value`` (the old value before it was updated)
    * ``change`` (the change in the value)
    * ``player_num`` (the player number the variable belongs to)

    For the ``change`` parameter, it will attempt to subtract the old value
    from the new value. If that works, it will return the result as the change.
    If it doesn't work (like if you're not storing numbers in this variable),
    then the change parameter will be *True* if the new value is different and
    *False* if the value didn't change.

    For examples, the following three lines:

    .. code::

        self.machine.player.score = 0
        self.machine.player.score += 500
        self.machine.player.score = 1200

    ... will cause the following three events to be posted:

    ``player_score`` with Args: ``value=0, change=0, prev_value=0``
    ``player_score`` with Args: ``value=500, change=500, prev_value=0``
    ``player_score`` with Args: ``value=1200, change=700, prev_value=500``

    """

    monitor_enabled = False
    """Class attribute which specifies whether any monitors have been registered
    to track player variable changes.
    """

    def __init__(self, machine, index):
        """Initialise player."""
        # use self.__dict__ below since __setattr__ would make these player vars
        self.__dict__['log'] = logging.getLogger("Player")
        self.__dict__['machine'] = machine
        self.__dict__['vars'] = dict()
        self.__dict__['_events_enabled'] = False

        number = index + 1

        self.log.debug("Creating new player: Player %s. (player index '%s')", number, index)

        # Set these after the player_added event so any player monitors
        # get notification of the new player before they start seeing variable
        # changes for it.
        self.vars['index'] = index
        '''player_var: index

        desc: The index of this player, starting with 0. For example, Player
        1 has an index of 0, Player 2 has an index of 1, etc.

        If you want to get the player number, use the "number" player variable
        instead.
        '''

        self.vars['number'] = number
        '''player_var: number

        desc: The number of the player, beginning with 1. (e.g. Player 1 has
        a number of "1", Player 2 is "2", etc.
        '''

        self._load_initial_player_vars()

        # Set the initial player score to 0
        self.__setattr__("score", 0)
        '''player_var: score

        desc: The player's score.
        '''

    def _load_initial_player_vars(self):
        """Load initial player var values from config."""
        if 'player_vars' not in self.machine.config:
            return

        config = self.machine.config['player_vars']
        for name, element in config.items():
            element = self.machine.config_validator.validate_config("player_vars", copy.deepcopy(element))
            self[name] = Util.convert_to_type(element['initial_value'], element['value_type'])

    def enable_events(self, enable=True, send_all_variables=True):
        """Enable/disable player variable events.

        Args:
        ----
            enable: Flag to enable/disable player variable events
            send_all_variables: Flag indicating whether or not to send an event
                with the current value of every player variable.
        """
        self._events_enabled = enable   # noqa

        # Send all current player variable values as events (if requested)
        if enable and send_all_variables:
            self.send_all_variable_events()

    def send_all_variable_events(self):
        """Send a player variable event for the current value of all player variables."""
        for name, value in self.vars.items():
            if isinstance(value, (int, str, float)):
                if isinstance(value, str):
                    self._send_variable_event(name, value, value, False, self.vars['number'])
                else:
                    self._send_variable_event(name, value, value, 0, self.vars['number'])

    # pylint: disable-msg=too-many-arguments
    def _send_variable_event(self, name: str, value, prev_value, change, player_num: int):
        """Send a player variable event performs any monitor callbacks if configured.

        :param name: The player variable name
        :param value: The new variable value
        :param prev_value: The previous variable value
        :param change: The change in value or True/False
        :param player_num: The player number this variable belongs to
        """
        self.machine.events.post('player_' + name,
                                 value=value,
                                 prev_value=prev_value,
                                 change=change,
                                 player_num=player_num)
        '''event: player_(name)
        config_section: player_vars
        class_label: player_var

        desc: Posted when simpler types of player variables are added or
        change value.

        The actual event has (var_name) replaced with the name of the
        player variable that changed. Some examples:

        * player_score
        * player_shot_upper_lit_hit

        Lots of things are stored in player variables, so there's no way to
        build a complete list of what all the options are here. Elsewhere
        in the documentation, if you see something that says it's stored in
        a player variable, that means you'll get this event when that
        player variable is created or is changed.

        Note that this event is only posted for simpler types of player
        variables, including player variables that are integers, floating
        point numbers, or strings. More complex player variables (lists,
        dicts, etc.) do not get this event posted.

        This event is posted for a single player variable changing, meaning
        if multiple player variables change at the same time, multiple
        events will be posted, one for each change.

        args:

        value: The new value of this player variable.

        prev_value: The previous value of this player variable, e.g. what
        it was before the current value.

        change: If the player variable just changed, this will be the
        amount of the change. If it's not possible to determine a numeric
        change (for example, if this player variable is a string), then
        this *change* value will be set to the boolean *True*.

        player_num: The player number this variable just changed for,
        starting with 1. (e.g. Player 1 will have *player_num=1*, Player 4
        will have *player_num=4*, etc.)

        '''

        # note the monitor is only called for simpler var changes
        if Player.monitor_enabled and "player" in self.machine.monitors:
            for callback in self.machine.monitors['player']:
                callback(name=name, value=value,
                         prev_value=prev_value, change=change,
                         player_num=player_num)

    def __repr__(self):
        """Return string representation."""
        try:
            return "<Player {}>".format(self.vars['number'])
        except KeyError:
            return '<Player (new)>'

    def __getattr__(self, name):
        """Return value of attribute or initialise it with 0 when it does not exist."""
        if name in self.vars:
            return self.vars[name]

        return 0

    def __setattr__(self, name, value):
        """Set value and post event to inform about the change."""
        # prevent events for internal variables
        if name in self.__dict__:
            self.__dict__[name] = value
            return

        new_entry = False
        prev_value = 0
        if name in self.vars:
            prev_value = self.vars[name]
        else:
            new_entry = True

        self.vars[name] = value

        try:
            change = value - prev_value
        except TypeError:
            change = prev_value != value

        if (change or new_entry) and isinstance(value, (int, str, float)):
            self.log.debug("Setting '%s' to: %s, (prior: %s, change: %s)",
                           name, self.vars[name], prev_value, change)

            if self._events_enabled:
                self._send_variable_event(name, self.vars[name], prev_value, change, self.vars['number'])

    def __getitem__(self, name):
        """Allow array get access."""
        return self.__getattr__(name)

    def __setitem__(self, name, value):
        """Allow array set access."""
        self.__setattr__(name, value)

    def __iter__(self):
        """Iterate all player vars."""
        for name, value in self.vars.items():
            yield name, value

    def is_player_var(self, var_name):
        """Check if player var exists.

        Args:
        ----
            var_name: String name of the player variable to test.

        Returns: *True* if the variable exists and *False* if not.

        """
        return var_name in self.vars



"""Test Switch Position Mixin."""
import os
from unittest import TestCase
from mpf.commands.create_config import Command
from unittest.mock import call, patch, mock_open

class TestCommandCreateConfig(TestCase):

    @patch('os.makedirs')
    def test_get_create_mode(self, os_makedirs):
        with patch("builtins.open", mock_open()) as mock_file:
            Command.create_mode_structure("testmode", os.path.normpath("/home/test/modes"), os.path.normpath("/home/test/"))
            mock_file.assert_called_with(os.path.normpath("/home/test/modes/testmode/config/testmode.yaml"), "w")
            expected_calls = [call(os.path.normpath("/home/test/modes")), call(os.path.normpath("/home/test/modes/testmode/config"))]
            os_makedirs.assert_has_calls(expected_calls)

    @patch('os.makedirs')
    def test_get_create_machine_config(self, os_makedirs):
        with patch("builtins.open", mock_open()) as mock_file:
            Command.create_machine_config_structure("testmachine", os.path.normpath("/home/testmachine"))

            with open(os.path.normpath("/home/testmachine/config/config.yaml"), "w") as handle_config:
                handle_config.write.assert_any_call("#config_version=5")

            mock_file.assert_any_call(os.path.normpath("/home/testmachine/config/config.yaml"), "w")

            mock_file.assert_any_call(os.path.normpath("/home/testmachine/tests/test_testmachine.yaml"), "w")

            expected_calls = [call(os.path.normpath("/home/testmachine")), call(os.path.normpath("/home/testmachine/config")), call(os.path.normpath("/home/testmachine/tests"))]
            os_makedirs.assert_has_calls(expected_calls)

    @patch('os.makedirs')
    def test_get_create_shows(self, os_makedirs):
        with patch("builtins.open", mock_open()) as mock_file:
            Command.create_show_structure("testshow", os.path.normpath("/home/test/shows"), os.path.normpath("/home/test/"))
            mock_file.assert_called_with(os.path.normpath("/home/test/shows/testshow.yaml"), "w")
            os_makedirs.assert_called_once_with(os.path.normpath("/home/test/shows"))






"""Interface for drivers."""
import abc
from collections import namedtuple
from typing import Any
from mpf.core.platform import DriverConfig


PulseSettings = namedtuple("PulseSettings", ["power", "duration"])
HoldSettings = namedtuple("HoldSettings", ["power"])


class DriverPlatformInterface(metaclass=abc.ABCMeta):

    """Interface for drivers in hardware platforms.

    DriverPlatformInterface is an abstract base class that should be overridden for all
    driver interface classes on supported platforms.  This class ensures the proper required
    methods are implemented to support driver operations in MPF.
    """

    __slots__ = ["number", "config"]

    def __init__(self, config: DriverConfig, number: "Any") -> None:
        """Initialise driver."""
        self.number = number    # type: Any
        self.config = config    # type: DriverConfig

    @abc.abstractmethod
    def pulse(self, pulse_settings: PulseSettings):
        """Pulse a driver.

        Pulse this driver for a pre-determined amount of time, after which
        this driver is turned off automatically. Note that on most platforms,
        pulse times are a max of 255ms. (Beyond that MPF will send separate
        enable() and disable() commands.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def enable(self, pulse_settings: PulseSettings, hold_settings: HoldSettings):
        """Enable this driver, which means it's held "on" indefinitely until it's explicitly disabled."""
        raise NotImplementedError

    @abc.abstractmethod
    def disable(self):
        """Disable the driver."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_board_name(self):
        """Return the name of the board of this driver."""
        raise NotImplementedError

    def __repr__(self):
        """Return board + number."""
        return "<Driver {} {} (config: {})>".format(self.get_board_name(), self.number, self.config)
